#include "ttl_p.h"

// -- static -------------------------------------------------------------
const int TtWindow::MSG_DEFAULT = 0x8000;

// -- message dispatch system --------------------------------------------
namespace {
  class TtWindowHashEntry {
    static const int TABLE_SIZE;
    static TtWindowHashEntry *hashtable;

  public:
    static TtWindow *find( HWND h_wnd );
    static void add( TtWindow *window );
    static void remove( TtWindow *window );

  private:
    TtWindow *window;
    TtWindowHashEntry *next;

    TtWindowHashEntry();
    TtWindowHashEntry( TtWindow *window );
  };
  
  const int TtWindowHashEntry::TABLE_SIZE = 157;
  TtWindowHashEntry *TtWindowHashEntry::hashtable = new TtWindowHashEntry[TABLE_SIZE];

  TtWindowHashEntry::TtWindowHashEntry() {
    this->window = NULL;
    this->next   = NULL;
  }
  TtWindowHashEntry::TtWindowHashEntry( TtWindow *window ) {
    this->window = window;
    this->next   = NULL;
  }

  void TtWindowHashEntry::add( TtWindow *window ) {
    int hash = ((int)window->get_h_wnd()) % TtWindowHashEntry::TABLE_SIZE;
    TtWindowHashEntry *current = &TtWindowHashEntry::hashtable[hash];
    while ( current->next != NULL ) {
      current = current->next;
    }
    current->next = new TtWindowHashEntry( window );
  }

  TtWindow *TtWindowHashEntry::find( HWND h_wnd ) {
    int hash = ((int)h_wnd) % TtWindowHashEntry::TABLE_SIZE;
    TtWindowHashEntry *current = TtWindowHashEntry::hashtable[hash].next;
    while ( current != NULL ) {
      if ( current->window->get_h_wnd() == h_wnd ) { return current->window; }
      current = current->next;
    }
    return NULL;
  }
  
  void TtWindowHashEntry::remove( TtWindow *window ) {
    int hash = ((int)window->get_h_wnd()) % TtWindowHashEntry::TABLE_SIZE;
    TtWindowHashEntry *current = &TtWindowHashEntry::hashtable[hash];
    while ( current->next != NULL ) {
      if ( current->next->window == window ) {
        TtWindowHashEntry *_tmp = current->next;
        current->next = current->next;
        delete _tmp;
        break;
      }
      current = current->next;
    }
  }
}

LRESULT CALLBACK TtWindow::_ttl_wnd_proc( HWND h_wnd, UINT msg, WPARAM w_param, LPARAM l_param ) {
  TtWindow *window = TtWindow::find_window( h_wnd );
  return ( window != NULL ) ?
    window->wnd_proc( msg, w_param, l_param ) : DefWindowProc( h_wnd, msg, w_param, l_param );
}

TtWindow *TtWindow::find_window( HWND h_wnd ) {
  return TtWindowHashEntry::find( h_wnd );
}

void TtWindow::init_common_controls() {
  InitCommonControls();
}

// override at sub class
DWORD  TtWindow::get_ex_style()   { return 0; }
DWORD  TtWindow::get_style()      { return 0; }
PCTSTR TtWindow::get_class_name() { return NULL; }
bool TtWindow::created() { return true; }

LRESULT TtWindow::wnd_proc( UINT msg, WPARAM w_param, LPARAM l_param ) {
  return ( this->wnd_proc_super == NULL ) ?
    DefWindowProc( this->h_wnd, msg, w_param, l_param )
      : CallWindowProc( wnd_proc_super, this->h_wnd, msg, w_param, l_param );
}

LRESULT TtWindow::wm_hscroll_recved( int code, int position, HWND h_wnd ) {
  code; position; h_wnd; // x}~̈
  return TtWindow::MSG_DEFAULT;
}

LRESULT TtWindow::wm_vscroll_recved( int code, int position, HWND h_wnd ) {
  code; position; h_wnd; // x}~̈
  return TtWindow::MSG_DEFAULT;
}

// -- instance -----------------------------------------------------------
TtWindow::TtWindow() {
  this->h_wnd = NULL;
  this->id = 0;
  this->h_instance = NULL;
  this->wnd_proc_super = NULL;
}

bool TtWindow::create( TtWindow *parent ) {
  return this->create( parent, 0, parent->h_instance, 0 );
}
bool TtWindow::create( TtWindow *parent, int id ) {
  return this->create( parent, id, parent->h_instance, 0 );
}
bool TtWindow::create( TtWindow *parent, int id, HINSTANCE h_instance ) {
  return this->create( parent, id, h_instance, 0 );
}
bool TtWindow::create( TtWindow *parent, int id, HINSTANCE h_instance, LPVOID lp_param ) {
  return this->_create( parent, id, h_instance, lp_param, true );
}
bool TtWindow::create_not_child( TtWindow *parent, int id, HINSTANCE h_instance, LPVOID lp_param ) {
  return this->_create( parent, id, h_instance, lp_param, false );
}
bool TtWindow::_create( TtWindow *parent, int id, HINSTANCE h_instance, LPVOID lp_param, bool child ) {
  this->parent = parent;
  this->h_instance = h_instance;
  this->id = id;
  
  this->h_wnd = CreateWindowEx(
    this->get_ex_style(),
    this->get_class_name(),
    NULL,
    ( parent != NULL && child ) ? ( this->get_style() | WS_CHILD ) : ( this->get_style() ),
    CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
    parent ? parent->get_h_wnd() : NULL,
    (HMENU)id,
    h_instance,
    lp_param );
  if ( this->h_wnd == NULL ) { return false; }

  TtWindowHashEntry::add( this );

  if ( !(this->created()) ) { return false; }

  return true;
}

int TtWindow::close() {
  return SendMessage( this->h_wnd, WM_CLOSE, 0, 0 );
}

TtWindow::~TtWindow() {
  TtWindowHashEntry::remove( this );
}


HWND      TtWindow::get_h_wnd()      { return this->h_wnd; }
int       TtWindow::get_id()         { return this->id; }
HINSTANCE TtWindow::get_h_instance() { return this->h_instance; }

TtWindow *TtWindow::get_parent() {
  return this->parent;
}


RECT TtWindow::get_rect() {
  RECT rect;
  GetWindowRect( this->h_wnd, &rect );
  return rect;
}

int TtWindow::get_pos_x() { return this->get_rect().left; }
int TtWindow::get_pos_y() { return this->get_rect().top; }

int TtWindow::get_width() {
  RECT rect = this->get_rect();
  return rect.right - rect.left;
}
int TtWindow::get_height() {
  RECT rect = this->get_rect();
  return rect.bottom - rect.top;
}


int TtWindow::show() {
  return ShowWindow( this->h_wnd, SW_SHOW );
}

int TtWindow::hide() {
  return ShowWindow( this->h_wnd, SW_HIDE );
}

int TtWindow::minimize() {
  return ShowWindow( this->h_wnd, SW_MINIMIZE );
}

int TtWindow::maximize() {
  return ShowWindow( this->h_wnd, SW_MAXIMIZE );
}

int TtWindow::restore() {
  return ShowWindow( this->h_wnd, SW_RESTORE );
}

int TtWindow::set_show_state( int state ) {
  return ShowWindow( this->h_wnd, state );
}


int TtWindow::set_position( int x, int y ) {
  RECT rect;
  GetWindowRect( this->h_wnd, &rect );
  return MoveWindow( this->h_wnd, x, y, rect.right - rect.left, rect.bottom - rect.top, TRUE );
}

int TtWindow::set_size( int width, int height ) {
  RECT rect;
  GetWindowRect( this->h_wnd, &rect );
  return MoveWindow( this->h_wnd, rect.left, rect.top, width, height, TRUE );
}

int TtWindow::set_pos_size( int x, int y, int width, int height ) {
  return MoveWindow( this->h_wnd, x, y, width, height, TRUE );
}

int TtWindow::set_rect( RECT rect ) {
  return MoveWindow( this->h_wnd, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, TRUE );
}


bool TtWindow::set_bottom() {
  return SetWindowPos( this->h_wnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
}

bool TtWindow::set_top() {
  return SetWindowPos( this->h_wnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
}

bool TtWindow::set_under( HWND wnd ) {
  return SetWindowPos( this->h_wnd, wnd, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
}


int TtWindow::set_font( HFONT h_font ) {
  return SendMessage( this->h_wnd, WM_SETFONT, (WPARAM)h_font, (LPARAM)true );
}


int TtWindow::set_forground() {
  return SetForegroundWindow( this->h_wnd );
}


int TtWindow::set_text( const char *str ) {
  return SetWindowText( this->h_wnd, str );
}

int TtWindow::get_text( char *buf, int buf_size ) {
  return GetWindowText( this->h_wnd, buf, buf_size );
}

int TtWindow::messageloop() {
  MSG msg;
  while ( GetMessage( &msg, NULL, 0, 0 ) ) {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }
  return msg.wParam;
}

int TtWindow::override_wnd_proc() {
  this->wnd_proc_super = (WNDPROC)GetWindowLong( this->h_wnd, GWL_WNDPROC );
  return SetWindowLong( this->h_wnd, GWL_WNDPROC, (LONG)( TtWindow::_ttl_wnd_proc ) );
}
